// Copyright 2022 onart@github. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
#include "../externals/glad/glad.h"
#include "yr_game.h"
#include "yr_sys.h"
#include "yr_graphics.h"
#include "yr_input.h"
#include "yr_audio.h"

#include "logger.hpp"
#include "yr_math.hpp"
#include "yr_bits.hpp"

#include "yr_constants.hpp"

#include "../externals/boost/predef/platform.h"
#include "../externals/boost/predef/compiler.h"

#include <thread>

#if BOOST_PLAT_ANDROID
#include <game-activity/native_app_glue/android_native_app_glue.h>
#endif

namespace onart{

    constexpr int LOCATE_TYPE = 2;

    // 각각 좌측 상단 좌표, 가로/세로입니다.
    static vec2 rects[200] = {
    { 0.206f, 0.274f }, { 0.107f, 0.206f },
    { 0.119f, 0.406f }, { 0.609f, 0.178f },
    { 0.080f, 0.903f }, { 0.401f, 0.662f },
    { 0.958f, 0.288f }, { 0.126f, 0.635f },
    { 0.023f, 0.861f }, { 0.469f, 0.464f },
    { 0.023f, 0.986f }, { 0.173f, 0.848f },
    { 0.875f, 0.726f }, { 0.327f, 0.453f },
    { 0.933f, 0.367f }, { 0.120f, 0.328f },
    { 0.497f, 0.628f }, { 0.194f, 0.677f },
    { 0.970f, 0.764f }, { 0.855f, 0.043f },
    { 0.496f, 0.869f }, { 0.203f, 0.703f },
    { 0.093f, 0.681f }, { 0.440f, 0.204f },
    { 0.974f, 0.344f }, { 0.896f, 0.882f },
    { 0.693f, 0.344f }, { 0.620f, 0.098f },
    { 0.282f, 0.453f }, { 0.206f, 0.577f },
    { 0.533f, 0.832f }, { 0.372f, 0.131f },
    { 0.730f, 0.184f }, { 0.509f, 0.659f },
    { 0.172f, 0.363f }, { 0.558f, 0.522f },
    { 0.967f, 0.576f }, { 0.213f, 0.210f },
    { 0.155f, 0.807f }, { 0.665f, 0.716f },
    { 0.117f, 0.472f }, { 0.582f, 0.294f },
    { 0.835f, 0.978f }, { 0.361f, 0.314f },
    { 0.872f, 0.916f }, { 0.478f, 0.699f },
    { 0.500f, 0.193f }, { 0.962f, 0.587f },
    { 0.646f, 0.243f }, { 0.826f, 0.080f },
    { 0.648f, 0.426f }, { 0.676f, 0.135f },
    { 0.742f, 0.091f }, { 0.249f, 0.477f },
    { 0.852f, 0.839f }, { 0.598f, 0.348f },
    { 0.520f, 0.726f }, { 0.760f, 0.458f },
    { 0.859f, 0.702f }, { 0.863f, 0.264f },
    { 0.302f, 0.409f }, { 0.537f, 0.479f },
    { 0.282f, 0.209f }, { 0.969f, 0.107f },
    { 0.284f, 0.712f }, { 0.881f, 0.845f },
    { 0.709f, 0.642f }, { 0.213f, 0.849f },
    { 0.667f, 0.284f }, { 0.674f, 0.867f },
    { 0.077f, 0.265f }, { 0.467f, 0.318f },
    { 0.609f, 0.134f }, { 0.096f, 0.512f },
    { 0.727f, 0.612f }, { 0.012f, 0.891f },
    { 0.785f, 0.039f }, { 0.159f, 0.673f },
    { 0.475f, 0.726f }, { 0.452f, 0.802f },
    { 0.188f, 0.818f }, { 0.679f, 0.046f },
    { 0.979f, 0.221f }, { 0.326f, 0.570f },
    { 0.016f, 0.709f }, { 0.994f, 0.124f },
    { 0.572f, 0.032f }, { 0.473f, 0.205f },
    { 0.209f, 0.251f }, { 0.728f, 0.265f },
    { 0.110f, 0.990f }, { 0.557f, 0.549f },
    { 0.358f, 0.735f }, { 0.137f, 0.257f },
    { 0.378f, 0.345f }, { 0.575f, 0.196f },
    { 0.401f, 0.456f }, { 0.417f, 0.503f },
    { 0.195f, 0.936f }, { 0.944f, 0.082f },
    { 0.280f, 0.439f }, { 0.626f, 0.772f },
    { 0.860f, 0.485f }, { 0.742f, 0.863f },
    { 0.516f, 0.306f }, { 0.148f, 0.798f },
    { 0.301f, 0.613f }, { 0.229f, 0.622f },
    { 0.338f, 0.376f }, { 0.168f, 0.604f },
    { 0.131f, 0.461f }, { 0.674f, 0.737f },
    { 0.253f, 0.686f }, { 0.074f, 0.447f },
    { 0.905f, 0.768f }, { 0.218f, 0.297f },
    { 0.186f, 0.642f }, { 0.822f, 0.470f },
    { 0.256f, 0.589f }, { 0.444f, 0.340f },
    { 0.987f, 0.619f }, { 0.040f, 0.208f },
    { 0.538f, 0.110f }, { 0.689f, 0.365f },
    { 0.703f, 0.819f }, { 0.391f, 0.559f },
    { 0.473f, 0.499f }, { 0.754f, 0.200f },
    { 0.084f, 0.617f }, { 0.159f, 0.632f },
    { 0.865f, 0.599f }, { 0.223f, 0.097f },
    { 0.919f, 0.359f }, { 0.496f, 0.089f },
    { 0.141f, 0.638f }, { 0.857f, 0.492f },
    { 0.906f, 0.131f }, { 0.961f, 0.745f },
    { 0.292f, 0.010f }, { 0.167f, 0.456f },
    { 0.257f, 0.872f }, { 0.477f, 0.706f },
    { 0.553f, 0.447f }, { 0.207f, 0.024f },
    { 0.169f, 0.562f }, { 0.286f, 0.089f },
    { 0.300f, 0.865f }, { 0.648f, 0.343f },
    { 0.919f, 0.355f }, { 0.822f, 0.025f },
    { 0.749f, 0.555f }, { 0.480f, 0.198f },
    { 0.258f, 0.569f }, { 0.352f, 0.233f },
    { 0.302f, 0.094f }, { 0.521f, 0.346f },
    { 0.295f, 0.456f }, { 0.169f, 0.611f },
    { 0.507f, 0.357f }, { 0.466f, 0.879f },
    { 0.350f, 0.708f }, { 0.361f, 0.267f },
    { 0.434f, 0.951f }, { 0.399f, 0.489f },
    { 0.611f, 0.401f }, { 0.711f, 0.553f },
    { 0.604f, 0.161f }, { 0.277f, 0.665f },
    { 0.705f, 0.327f }, { 0.250f, 0.979f },
    { 0.561f, 0.547f }, { 0.801f, 0.371f },
    { 0.655f, 0.254f }, { 0.970f, 0.253f },
    { 0.968f, 0.725f }, { 0.598f, 0.907f },
    { 0.001f, 0.064f }, { 0.141f, 0.990f },
    { 0.979f, 0.405f }, { 0.787f, 0.769f },
    { 0.590f, 0.058f }, { 0.545f, 0.617f },
    { 0.506f, 0.344f }, { 0.111f, 0.147f },
    { 0.652f, 0.705f }, { 0.373f, 0.752f },
    { 0.679f, 0.414f }, { 0.408f, 0.255f },
    { 0.168f, 0.986f }, { 0.777f, 0.020f },
    { 0.749f, 0.302f }, { 0.089f, 0.813f },
    { 0.326f, 0.230f }, { 0.040f, 0.474f },
    { 0.431f, 0.139f }, { 0.926f, 0.918f },
    { 0.649f, 0.009f }, { 0.924f, 0.167f },
    { 0.073f, 0.881f }, { 0.518f, 0.569f },
    };

    const std::chrono::steady_clock::time_point longTp = std::chrono::steady_clock::now();
    float Game::_dt = 0.016f, Game::_idt = 60.0f;
    uint64_t Game::_tp = (longTp - std::chrono::steady_clock::time_point()).count();
    const float &Game::dt(Game::_dt), &Game::idt(Game::_idt);
    const uint64_t& Game::tp(Game::_tp);
    int32_t Game::_frame = 1;
    const int32_t& Game::frame(Game::_frame);
    YRGraphics* Game::vk = nullptr;
    Window* Game::window = nullptr;
    void* Game::hd = nullptr;

#ifndef YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD
#define YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD (BOOST_PLAT_ANDROID)
#else
#undef YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD
#define YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD 1
#endif

    static std::mutex eventQMutex;

    enum class WindowEvent{ WE_SIZE = 0, WE_KEYBOARD = 1, WE_CLICK = 2, WE_CURSOR = 3, WE_SCROLL = 4 };
    struct EV{
        WindowEvent sType;
        union{
            struct{int sizeX, sizeY;};
            struct{int key, scancode, action, mods;};
            struct{int mouseKey, mouseAction, mouseMods;};
            struct{double mouseX, mouseY;};
            struct{double scrollX, scrollY;};
        };
    };

    static std::vector<EV> eventQ;
    static std::vector<EV> tempQ;

    static void recordSizeEvent(int x, int y) {
        std::unique_lock<std::mutex> _(eventQMutex);
        eventQ.push_back({});
        EV& back = eventQ.back();
        back.sType = WindowEvent::WE_SIZE;
        back.sizeX = x;
        back.sizeY = y;
    }

    static void recordKeyEvent(int key, int scancode, int action, int mods) {
        std::unique_lock<std::mutex> _(eventQMutex);
        eventQ.push_back({});
        EV& back = eventQ.back();
        back.sType = WindowEvent::WE_KEYBOARD;
        back.key = key;
        back.scancode = scancode;
        back.action = action;
        back.mods = mods;
    }

    static void recordClickEvent(int key, int action, int mods) {
        std::unique_lock<std::mutex> _(eventQMutex);
        eventQ.push_back({});
        EV& back = eventQ.back();
        back.sType = WindowEvent::WE_CLICK;
        back.mouseKey = key;
        back.mouseAction = action;
        back.mouseMods = mods;
    }

    static void recordCursorEvent(double x, double y) {
        std::unique_lock<std::mutex> _(eventQMutex);
        eventQ.push_back({});
        EV& back = eventQ.back();
        back.sType = WindowEvent::WE_CURSOR;
        back.mouseX = x;
        back.mouseY = y;
    }

    static void recordScrollEvent(double x, double y) {
        std::unique_lock<std::mutex> _(eventQMutex);
        eventQ.push_back({});
        EV& back = eventQ.back();
        back.sType = WindowEvent::WE_SCROLL;
        back.scrollX = x;
        back.scrollY = y;
    }

    bool loaded = false; // async 테스트용
    static float thr = 0.5f;
    
    int Game::start(void* hd, Window::CreationOptions* opt){
        if(window) {
            LOGWITH("Warning: already started");
            return 2;
        }

        Game::hd = hd;

        delete window;        
        window = new Window(hd, opt);
        if(!window->isNormal()) {
            delete window;
            LOGWITH("Window creation failed");
            Window::terminate();
            return 1;
        }

#if !YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD
        std::thread gamethread([]() {
#endif
            window->setMainThread();
            vk = new YRGraphics(window);
            if (!init()) {
                delete window;
                Window::terminate();
                return;
            }
            for (;; _frame++) {
                pollEvents();
                if (window->windowShouldClose()) break;
                std::chrono::duration<uint64_t, std::nano> longDt = std::chrono::steady_clock::now() - longTp;
                // [0, 2^52 - 1] 범위 정수를 균등 간격의 [1.0, 2.0) 범위 double로 대응시키는 함수에 십억을 적용한 후 1을 뺀 결과에 곱하여 1로 만들 수 있는 수
                constexpr double ONE_SECOND = 1.0 / 0.0000002220446049250313080847263336181640625;
                
                uint64_t ndt = longDt.count() - _tp;
                double ddt = (fixedPointConversion64i(ndt) - 1.0) * ONE_SECOND; // 함수를 사용할 수 없는 ndt의 상한선: 4,503,599,627,370,496 > 52일
                double iddt = 1.0 / ddt;
                _tp = longDt.count();
                _dt = static_cast<float>(ddt);
                _idt = static_cast<float>(iddt);
                YRGraphics::handle();
                auto rp2s = YRGraphics::getRenderPass2Screen(0);
                auto vb = YRGraphics::getMesh(0);
                auto tx = YRGraphics::getTexture(0);
                auto ubo = YRGraphics::getUniformBuffer(0);
                int x, y;
                static double fts = 0;
                if ((frame & 255) == 0) { // 256프레임마다 1회 평균값 출력
                    printf("%f\r", 255.0 / fts);
                    fts = 0; // printf를 수행한 프레임은 제외됨. 그래서 255 / 프레임타임 누적값
                }
                else {
                    fts += ddt;
                }
                window->getFramebufferSize(&x, &y);
                float aspect = 1.0f;// (float)x / y;
                float pushed = PI<float> / 2;// std::abs(std::sin((double)_tp * 0.000000001));
                if (vk->swapchain.handle) {
                    
                    rp2s->start();

                    rp2s->bind(0, tx);
                    vec2 size((float)x, (float)y);
                    for (int i = 0; i < 100; i++) {
                        if constexpr (LOCATE_TYPE == 0) {
                            vec2 sz = rects[2 * i + 1] * size;
                            vec2 lt = rects[2 * i] * size;
                            rp2s->setViewport(sz.x, sz.y, lt.x, lt.y, true);
                            rp2s->invoke(vb);
                        }
                        else if constexpr (LOCATE_TYPE == 1) {
                            vec4 pushed(rects[2 * i].xyyy());
                            pushed.z = rects[2 * i + 1].x;
                            pushed.w = rects[2 * i + 1].y;
                            rp2s->push(pushed.entry, 0, 16);
                            rp2s->invoke(vb);
                        }
                        else if constexpr(LOCATE_TYPE == 2){
                            vec4 verts[6];
                            const vec2 pos[6] = { vec2(0, -1), vec2(0, 0), vec2(1, -1), vec2(1, 0), vec2(1, -1), vec2(0, 0) };
                            for (int j = 0; j < 6; j++) {
                                vec2 transformed = (pos[j] * rects[2 * i + 1] + vec2(rects[2 * i].x, -rects[2 * i].y)) * 2.0f - vec2(1.0f, -1.0f);
                                verts[j].x = transformed.x;
                                verts[j].y = transformed.y;
                                verts[j].z = pos[j].x;
                                verts[j].w = -pos[j].y;
                            }
                            vb->update(verts, 0, sizeof(verts));
                            rp2s->invoke(vb);
                        }
                    }

                    rp2s->execute();
                    
                    if constexpr (YRGraphics::OPENGL_GRAPHICS) {
                        glfwSwapBuffers((GLFWwindow*)window->window);
                    }
                }
            }
#if !YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD
        });
        while(!window->windowShouldClose()) { window->waitEvents(); }
        gamethread.join();
#endif

        finalize();
        return 0;
    }

    void Game::pollEvents(){
#if YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD
        window->pollEvents();
#else
        eventQMutex.lock();
        tempQ.swap(eventQ);
        eventQMutex.unlock();
        int sx = -1, sy = -1;
        double cx = -1, cy = -1;
        for(EV& ev: tempQ) {
            switch(ev.sType){
                case WindowEvent::WE_SIZE:
                    sx = ev.sizeX;
                    sy = ev.sizeY;
                    break;
                case WindowEvent::WE_KEYBOARD:
                    Input::keyboard(ev.key, ev.scancode, ev.action, ev.mods);
                    break;
                case WindowEvent::WE_CLICK:
                    Input::click(ev.mouseKey, ev.mouseAction, ev.mouseMods);
                    break;
                case WindowEvent::WE_CURSOR:
                    cx = ev.mouseX;
                    cy = ev.mouseY;
                    break;
                case WindowEvent::WE_SCROLL:
                    // TODO. ev.scrollX, ev.scrollY
                    break;
            }
        }
        tempQ.clear();
        if(sx != -1) {
            windowResized(sx, sy);
        }
        if(cx != -1) {
            Input::moveCursor(cx, cy);
        }
#endif
    }

    void Game::finalize(){
        Audio::finalize();
        delete vk;
        delete window; // Window보다 스왑체인을 먼저 없애야 함 (안 그러면 X11에서 막혀서 프로그램이 안 끝남)
        window = nullptr;
        vk = nullptr;
        Window::terminate();
    }

    bool Game::init() {
        Audio::init();
#if YR_NO_NEED_TO_USE_SEPARATE_EVENT_THREAD
        window->clickCallback = Input::click;
        window->keyCallback = Input::keyboard;
        window->posCallback = Input::moveCursor;
        window->touchCallback = Input::touch;
        window->windowSizeCallback = windowResized;
#else
        window->clickCallback = recordClickEvent;
        window->keyCallback = recordKeyEvent;
        window->posCallback = recordCursorEvent;
        window->windowSizeCallback = recordSizeEvent;
        window->scrollCallback = recordScrollEvent;
#endif

        auto rtt = YRGraphics::RenderTargetType::COLOR1;
        auto rp2s = YRGraphics::createRenderPass2Screen(nullptr, 1, 0, false);
        using testv_t = YRGraphics::Vertex<vec3, vec2>;
        testv_t verts[]{ {{-1,-1,0},{0,0}},{{-1,1,0},{0,1}},{{1,-1,0},{1,0}},{{1,1,0},{1,1}} };
        uint16_t inds[]{ 0,1,2,2,1,3 };
#ifdef YR_USE_VULKAN
        if constexpr (YRGraphics::VULKAN_GRAPHICS) {
            auto texLayout = YRGraphics::getTextureLayout(0);
            VkPipelineLayout lo{};
            auto fs = YRGraphics::createShader(COMMON_FRAG, sizeof(COMMON_FRAG), 1);
            VkShaderModule vs{};
            if constexpr (LOCATE_TYPE == 0) {
                vs = YRGraphics::createShader(VP_VS, sizeof(VP_VS), 0);
                lo = YRGraphics::createPipelineLayout(&texLayout, 1, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 0);
            }
            else if constexpr (LOCATE_TYPE == 1) {
                vs = YRGraphics::createShader(PUSH_VS, sizeof(PUSH_VS), 0);
                lo = YRGraphics::createPipelineLayout(&texLayout, 1, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 0);
            }
            else if constexpr(LOCATE_TYPE == 2){
                vs = YRGraphics::createShader(DUV_VS, sizeof(DUV_VS), 0);
                auto ub = YRGraphics::createUniformBuffer(100, 16, VK_SHADER_STAGE_VERTEX_BIT, 0, 1);
                VkDescriptorSetLayout layouts[] = {texLayout, ub->getLayout()};
                lo = YRGraphics::createPipelineLayout(layouts, 2, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 0);
            }
            YRGraphics::createPipeline(nullptr, 0, 0, nullptr, 0, 0, rp2s, 0, 0, lo, vs, fs, 0);
            YRGraphics::createNullMesh(6, 0);
            loaded = true; YRGraphics::createTexture(TEX0, sizeof(TEX0), 4, 0, YRGraphics::isSurfaceSRGB());
        }
#elif defined(YR_USE_OPENGL)
        const char TEST_GL_VERT1[] = R"(
#version 450

layout(location = 0) in vec3 inPosition;
layout(location = 1) in vec2 inTc;

layout(location = 0) out vec2 tc;

layout(std140, binding=11) uniform ui{
    mat4 aspect;
    float t;
};

void main() {
    gl_Position = aspect * vec4(inPosition, 1.0);
    tc = inTc;
}
)";

        const char TEST_GL_FRAG1[] = R"(
#version 450

layout(location = 0) in vec2 tc;

out vec4 outColor;
layout(binding = 0) uniform sampler2D tex;

layout(std140, binding=11) uniform ui{
    mat4 aspect;
    float t;
};

void main() {
    outColor = texture(tex, tc);
}
)";
        if constexpr (YRGraphics::OPENGL_GRAPHICS) {
            auto vs = YRGraphics::createShader(TEST_GL_VERT1, sizeof(TEST_GL_VERT1), 0, YRGraphics::ShaderType::VERTEX);
            auto fs = YRGraphics::createShader(TEST_GL_FRAG1, sizeof(TEST_GL_FRAG1), 1, YRGraphics::ShaderType::FRAGMENT);
            YRGraphics::PipelineInputVertexSpec sp[2];
            testv_t::info(sp);
            auto pp = YRGraphics::createPipeline(sp,sizeof(testv_t),2,nullptr, 0, 0, vs, fs, 0);
            offrp->usePipeline(pp, 0);
            offrp->usePipeline(pp, 1);
            rp2s->usePipeline(pp, 0);
            //YRGraphics::asyncCreateTexture(TEX0, sizeof(TEX0), 4, [](variant8) { loaded = true; }, 0, YRGraphics::isSurfaceSRGB(), true, false);
            YRGraphics::createTexture(TEX0, sizeof(TEX0), 4, 0, YRGraphics::isSurfaceSRGB()); loaded = true;
            YRGraphics::createNullMesh(3, 1);
            YRGraphics::createMesh(verts, sizeof(testv_t), 4, inds, 2, 6, 0);
        }
#elif defined(YR_USE_D3D11)
        const char VP_VS[] = R"(
struct VS_INPUT{
    float _: _0_;
    uint vid: SV_VertexID;
};

struct PS_INPUT{
    float4 pos: SV_POSITION;
    float2 tc: TEXCOORD0;
};

static const float2 pos[6] = {float2(-1, -1), float2(-1, 1), float2(1, -1), float2(1, 1), float2(1, -1), float2(-1, 1)};

PS_INPUT main(VS_INPUT input) {
    PS_INPUT ret = (PS_INPUT)0;
    ret.pos = float4(pos[input.vid], 0.0, 1.0);
    ret.tc = max(pos[input.vid],float2(0.0,0.0));
    ret.tc.y = 1-ret.tc.y;
    return ret;
}
)";

        const char CB_VS[] = R"(
struct VS_INPUT{
    float _: _0_;
    uint vid: SV_VertexID;
};

cbuffer _0: register(b13){
    float2 offset;
    float2 zoom;
};

const static float2 pos[6] = {float2(0, -1), float2(0, 0), float2(1, -1), float2(1, 0), float2(1, -1), float2(0, 0)};

struct PS_INPUT{
    float4 pos: SV_POSITION;
    float2 tc: TEXCOORD0;
};

PS_INPUT main(VS_INPUT input) {
    PS_INPUT ret = (PS_INPUT)0;

    ret.pos = float4((pos[input.vid] * zoom + float2(offset.x,-offset.y)) * 2.0 - float2(1.0,-1.0), 0.0, 1.0);
    ret.tc = pos[input.vid] + float2(0,1);
    ret.tc.y = 1.0 - ret.tc.y;
    return ret;
}
)";

    const char VB_VS[] = R"(
struct VS_INPUT{
    float4 xyuv: _0_;
};

struct PS_INPUT{
    float4 pos: SV_POSITION;
    float2 tc: TEXCOORD0;
};

PS_INPUT main(VS_INPUT input) {
    PS_INPUT ret = (PS_INPUT)0;

    ret.pos = float4(input.xyuv.xy, 0, 1);
    ret.tc = input.xyuv.zw;
    return ret;
}
)";

        const char TEST_D11_VERT1[] = R"(
struct VS_INPUT{
    float3 inPosition: _0_;
    float2 inTc: _1_;
};

struct PS_INPUT{
    float4 pos: SV_POSITION;
    float2 tc: TEXCOORD0;
};

cbuffer _0: register(b13){
    float4x4 aspect;
    float t;
}

PS_INPUT main(VS_INPUT input) {
    PS_INPUT ret = (PS_INPUT)0;
    ret.pos = mul(float4(input.inPosition, 1.0), aspect);
    ret.tc = input.inTc;
    return ret;
}
)";

        const char TEST_D11_FRAG1[] = R"(

struct PS_INPUT{
    float4 pos: SV_POSITION;
    float2 tc: TEXCOORD0;
};

Texture2D tex: register(t0);
SamplerState spr: register(s0);

float4 main(PS_INPUT input): SV_TARGET {
    return tex.Sample(spr, input.tc);
}
)";
        ID3DBlob* vsb{};
        ID3DBlob* psb = compileShader(TEST_D11_FRAG1, sizeof(TEST_D11_FRAG1), YRGraphics::ShaderType::FRAGMENT);
        YRGraphics::PipelineInputVertexSpec desc;
        int vsize = 4;
        auto ps = YRGraphics::createShader(psb->GetBufferPointer(), psb->GetBufferSize(), 1, YRGraphics::ShaderType::FRAGMENT);
        ID3D11DeviceChild* vs{};
        if constexpr (LOCATE_TYPE == 0) {
            vsb = compileShader(VP_VS, sizeof(VP_VS), YRGraphics::ShaderType::VERTEX);
            vs = YRGraphics::createShader(vsb->GetBufferPointer(), vsb->GetBufferSize(), 0, YRGraphics::ShaderType::VERTEX);
            YRGraphics::createNullMesh(6, 0);
            vsize = 4;
            YRGraphics::Vertex<float>::info(&desc, 0);
        }
        else if constexpr (LOCATE_TYPE == 1) {
            vsb = compileShader(CB_VS, sizeof(CB_VS), YRGraphics::ShaderType::VERTEX);
            vs = YRGraphics::createShader(vsb->GetBufferPointer(), vsb->GetBufferSize(), 0, YRGraphics::ShaderType::VERTEX);
            YRGraphics::createNullMesh(6, 0);
            vsize = 4;
            YRGraphics::Vertex<float>::info(&desc, 0);
        }
        else if constexpr (LOCATE_TYPE == 2) {
            vsb = compileShader(VB_VS, sizeof(VB_VS), YRGraphics::ShaderType::VERTEX);
            vs = YRGraphics::createShader(vsb->GetBufferPointer(), vsb->GetBufferSize(), 0, YRGraphics::ShaderType::VERTEX);
            vsize = 16;
            YRGraphics::Vertex<vec4>::info(&desc, 0);
            YRGraphics::createMesh(nullptr, 16, 6, nullptr, 2, 0, 0, false);
        }
        
        auto pp = YRGraphics::createPipeline(&desc, vsize, 1, nullptr, 0, 0, vsb->GetBufferPointer(), vsb->GetBufferSize(), vs, ps, 0, false);
        rp2s->usePipeline(pp, 0);
        //YRGraphics::asyncCreateTexture(TEX0, sizeof(TEX0), 4, [](variant8) { loaded = true; }, 0, YRGraphics::isSurfaceSRGB(), true, false);
        YRGraphics::createTexture(TEX0, sizeof(TEX0), 4, 0, false); loaded = true;
        vsb->Release();
        psb->Release();
#endif
        return true;
    }

    void Game::windowResized(int x, int y){
#if BOOST_PLAT_ANDROID
        vk->createSwapchain(x, y, window);
#else
        vk->createSwapchain(x, y);
#endif
    }

    void Game::exit(){
        window->close();
    }

    void Game::readFile(const char* fileName, std::basic_string<uint8_t>* buffer){
        buffer->clear();
#if BOOST_PLAT_ANDROID
        AAsset* asset = AAssetManager_open(reinterpret_cast<android_app*>(hd)->activity->assetManager, fileName, AASSET_MODE_BUFFER);
        if(asset) {
            int len = AAsset_getLength(asset);
            buffer->resize(len);
            AAsset_read(asset, buffer->data(), len);
            AAsset_close(asset);
        }
#else
        FILE* fp = fopen(fileName, "rb");
        if(fp) {
            fseek(fp, 0, SEEK_END);
            int len = ftell(fp);
            buffer->resize(len);
            fseek(fp, 0, SEEK_SET);
            fread(buffer->data(), 1, len, fp);
            fclose(fp);
        }
#endif
    }
}